/*
 * 2D Polygon Landscape
 *
 * Copyright (c) 2011  Tobias Zwick
 */

#include <map>
#include <iostream>

#define _USE_MATH_DEFINES
#include <cmath>

#include "Line.h"

/* --------------------- Line 2D ----------------------*/

Line::Line() : _A(0,0), _B(1,0)
{
	type = Normal;
	line = Horizontal;
}

Line::Line(const Vertex &first, const Vertex &second, Line_type typ) : _A(first), _B(second)
{
	type = typ;
	updateDir();	
}

Line::Line(const Line& l) : _A(l.A()), _B(l.B())
{
	type = l.getType();
	updateDir();
}

bool Line::operator==(const Line& l) const
{
	return _A == l.A() && _B == l.B() && type == l.type && line == l.line;
}

bool Line::operator!=(const Line& l) const
{
	return !(*this == l);
}

Line Line::operator+() const
{
	return Line(*this);
}

Line Line::operator-() const
{
	// special for ray: turn 180° around but keep the same starting point
	if(type == Ray) return Line(_A,-_B+_A*2,type);
	// otherwise, just switch A with B (direction vector)
	return Line(_B,_A,type);
}

void Line::updateDir()
{
	// is a line parallel to axis? -> simpler to calculate
	if(_A.x() == _B.x()) line = Vertical;
	else if(_A.y() == _B.y()) line = Horizontal;
	else line = Standard;

}

bool Line::intersects(const Line& other) const
{
	Intersection intersect;
	return intersects(other,intersect);
}

bool Line::intersects(const Line& other, Intersection &intersect) const
{
	/*
	Returns false if there is no intersection. If there is, returns true
	and returns one intersection. If either this or the other line is
	parallel to one axis, the calculation is simpler (about half as
	expensive).

	This is how an intersection is calculated:

	A+X(B-A) is the this line, P+Y(Q-P) the other.
	X and Y are not vectors but scalars.

	A+X(B-A) = P+Y(Q-P)              # equate both lines
	
	A.x+X(B.x-A.x) = P.x+Y(Q.x-P.x)  # break it up in x and y
	A.y+X(B.y-A.y) = P.y+Y(Q.y-P.y)

	X = ((P.x-A.x)+Y(Q.x-P.x)) / (B.x-A.x)   # to X
	X = ((P.y-A.y)+Y(Q.y-P.y)) / (B.y-A.y)              

	X = (PA.x+Y*QP.x) / BA.x                 # sum B-A to BA etc...
	X = (PA.y+Y*QP.y) / BA.y	

	(PA.x+Y*QP.x) / BA.x = (PA.y+Y*QP.y) / BA.y	# equate

	(PA.x+Y*QP.x) * BA.y = (PA.y+Y*QP.y) * BA.x

	BA.y * PA.x + Y * QP.x * BA.y = PA.y * BA.x + Y*QP.y * BA.x

	BA.y * PA.x + Y*QP.x*BA.y = PA.y * BA.x + Y*QP.y*BA.x

	Y*QP.x*BA.y - Y*QP.y*BA.x = PA.y * BA.x - BA.y * PA.x

	Y*(QP.x*BA.y - QP.y*BA.x) = PA.y*BA.x - BA.y*PA.x

	Y = (PA.y*BA.x - BA.y*PA.x) / (QP.x*BA.y - QP.y*BA.x)

	  if the denominator is zero, the lines are parallel. If the numerator
	  is zero, too, the lines are on the same line.
	  If Y it is not within it's bonds, there is no intersection. (The same
	  applies for X)

	  One of them might be divisin by zero. We can use the other one.
	  Both only could be /0 if this line is a point:

	X = (PA.x+Y*QP.x)/BA.x
	X = (PA.y+Y*QP.y)/BA.y

	*/

	// its a point.....
	if(_A == _B) {
		double val;
		if(other.contains(_A,val)) {
			intersect.clear();
			intersect.point = new Vertex(_A);
			intersect.line = NULL;
			intersect.thisParam = 0.0;
			intersect.otherParam = val;
			intersect.otherNormal = other.getNormal();
			intersect.exact = true;
			return true;
		}
	}
	// parallel to axis
	if(line != Standard) return simpleIntersect(other, intersect);
	if(other.line != Standard)
	{
		if(other.simpleIntersect(*this, intersect))
		{
			// switch paramValue
			double thisParam = intersect.thisParam;
			intersect.thisParam = intersect.otherParam;
			intersect.otherParam = thisParam;
			Vertex thisNormal = intersect.thisNormal;
			intersect.thisNormal = intersect.otherNormal;
			intersect.otherNormal = thisNormal;
			// because we used the intersect() from the other line
			return true;
		}
		return false;
	}

	Vertex BA = getVector();
	Vertex PA = other.A()-_A;
	Vertex QP = other.B()-other.A();

	double denominator = (double) (BA.y()*QP.x() - BA.x()*QP.y());
	double numerator = (double) (BA.x()*PA.y() - BA.y()*PA.x());
	// parallel!
	if(denominator == 0)
	{
		if(numerator != 0) return false; 
		// on same line
		else return sameLineIntersect(other,intersect);
	}

	// in parameter bounds?
	double Y = numerator/denominator;
	if(!other.paramBounds(Y)) return false;

	// we don't want division by zero...
	if(BA.y() != 0) {
		numerator = (double) (PA.y() + Y*QP.y());
		denominator = (double) BA.y();
	}
	else if(BA.x() != 0) {
		numerator = (double) (PA.x() + Y*QP.x());
		denominator = (double) BA.x();
	}

	// in parameter bounds
	double X =  numerator / denominator;
	if(!paramBounds(X)) return false;

	// this is the point...
	double vx = _A.x()+X*BA.x();
	double vy = _A.y()+X*BA.y();
	intersect.point = new Vertex((int)floor(vx+0.5), (int)floor(vy+0.5));
	intersect.exact = (floor(vx) == ceil(vx) && floor(vy) == ceil(vy));

	// fill the values
	intersect.thisParam = X;
	intersect.otherParam = Y;

	intersect.thisNormal = getNormal();
	if(getDirection(other.A()) < 0)
		intersect.thisNormal = -getNormal();

	intersect.otherNormal = other.getNormal();
	if(other.getDirection(_A) < 0)
		intersect.otherNormal = -other.getNormal();

	return true;
}

bool Line::simpleIntersect(const Line &other, Intersection &intersect) const
{ 
	/*
	This is the intersection with a line that is parallel to one axis
	This may not look a lot easier but actually it is about 2x simpler,
	his might make a difference with lots of intersections to check:

	If this line goes only on the Y-Axis, we can equate like this
	
	A.x = P.x+Y*(Q.x-P.x)              # equate both lines

	Y*(Q.x-P.x) = A.x - P.x

	Y = (A.x - P.x)/(Q.x - P.x)        # to Y

	If this line goes on the X-Axis, then its the same with #.y

	X = (PA.x+Y*QP.x)/BA.x
	X = (PA.y+Y*QP.y)/BA.y

	*/

	double denominator;
	double numerator;

	Vertex QP = other.getVector();

	// line goes on Y-Axis
	if(line == Vertical)
	{
		numerator = (double) (_A.x() - other.A().x());
		denominator = (double) QP.x();
	}
	// line goes on X-Axis
	else if(line == Horizontal)
	{
		numerator = (double) (_A.y() - other.A().y());
		denominator = (double) QP.y();
	}

	// lines are parallel
	if(denominator == 0)
	{
		// not on same line
		if(numerator != 0) return false; 
		else return sameLineIntersect(other, intersect);
	}

	double Y =  numerator / denominator;
	if(!other.paramBounds(Y)) return false;

	Vertex BA = getVector();
	Vertex PA = other.A() - _A;

	double X;

	if(line == Vertical)
		X = (double) (PA.y() + Y*QP.y()) / (double) BA.y();
	else if(line == Horizontal)
		X = (double) (PA.x() + Y*QP.x()) / (double) BA.x();

	if(!paramBounds(X)) return false;

	// calculate the point
	double vx = other.A().x() + Y*QP.x();
	double vy = other.A().y() + Y*QP.y();
	intersect.clear();
	intersect.point = new Vertex((int)floor(vx+0.5), (int)floor(vy+0.5));
	intersect.exact = (floor(vx) == ceil(vx) && floor(vy) == ceil(vy));

	// and return...
	intersect.thisParam = X;
	intersect.otherParam = Y;

	intersect.thisNormal = getNormal();
	if(getDirection(other.A()) < 0)
		intersect.thisNormal = -getNormal();

	intersect.otherNormal = other.getNormal();
	if(other.getDirection(_A) < 0)
		intersect.otherNormal = -other.getNormal();

	return true;
}

bool Line::sameLineIntersect(const Line &other, Intersection &intersect) const
{
	/*
	Now we got the infnormation that B-A and Q-P go into the same (or
	opposite) direction.
	How to find out if they intersect (e.g. for line segments):

	For both lines, look if the other two points are contained in the line.
	*/

	std::map<double,Vertex> p;

	double pv;
	if(contains(other.A(),pv)) p[pv] = other.A();
	if(contains(other.B(),pv)) p[pv] = other.B();
	if(other.contains(A())) p[0.0] = A();
	if(other.contains(B())) p[1.0] = B();

	// in the set there can be maximum 2 different points and minimum zero
	if(p.empty())
	{
		return false;
	}
	else
	{
		intersect.clear();
		intersect.thisNormal = getNormal();
		intersect.otherNormal = other.getNormal();

		// the lines share only one point
		if(p.size() == 1)
		{
			intersect.point = new Vertex( p.begin()->second );
			intersect.exact = true;
			contains(p.begin()->second,intersect.thisParam);
			other.contains(p.begin()->second,intersect.otherParam);
			return true;
		}
		// line
		else if(p.size() >= 2)
		{
			Vertex P1(p.begin()->second);
			Vertex P2((--(p.end()))->second);
			Line_type ntype = Normal;

			// if any of the two lines is a line segment, the intersection is also a
			// line segment
			if(getType() == Segment || other.getType() == Segment)
			{
				ntype = Segment;
			}
			// otherwise
			else
			{
				bool oppositeDirection = getVector()*other.getVector() < 0;

				// the intersection of two rays is a line segment if the
				// rays have opposite direction
				if(getType() == Ray && other.getType() == Ray)
				{
					if(oppositeDirection)
						ntype = Segment;
					else
						ntype = Ray;
				}
				else if(getType() == Ray)
				{
					ntype = Ray;
				}
				else if(other.getType() == Ray)
				{
					ntype = Ray;
					// the direction vector of the intersection is always the same as the
					// direction vector of this line, except if the direction has a
					// meaning: the other line is a ray and goes into the opposite
					// direction as this line.
					if(oppositeDirection)
					{
						P2 = p.begin()->second;
						P1 = (--(p.end()))->second;
					}
				}
			}

			intersect.line = new Line(P1,P2,ntype);
			intersect.exact = true;
			// just of point one...
			// the correct way would be to look which of the two points is closer to A
			// of each line
			contains(P1,intersect.thisParam);
			other.contains(P1,intersect.otherParam);

			return true;
		}
	}

	return false;
}

inline bool Line::paramBounds(double param) const
{
	// always inside bounds for lines
	if(type == Normal) return true;

	bool result;
	// for others, check if the parameter is bigger or equal 0 and 
	// for the line segment, smaller or equal one
	if(type == Ray) result = (param >= 0.0);
	if(type == Segment) result = (param >= 0.0 && param <= 1.0);

	return result;
}

bool Line::contains(const Vertex& point) const
{
	double paramVal;
	return contains(point, paramVal);
}

bool Line::contains(const Vertex& point, double &paramVal) const
{
	/*
	Checks if the point is contained in the line. The calculation:

	A + x*(B-A) = P
	
	x*(B-A) = P-A
	
	x*BA = PA
	
	x*BAx = PAx
	x*BAy = PAy
	
	x = PAx / BAx
	x = PAy / BAy
	
	PAx / BAx = PAy / BAy
	
	if BAx is zero, PAx must be zero too. The same for BAy & PAy. If that is
	the case, the value of x1 is only set by the other term

	And for rays and line segments, the value for x must be checked

	*/

	Vertex BA = getVector();
	Vertex PA = point-_A;

	double x1, x2;

	if(BA.x() != 0)
	{
		x1 = (double) PA.x() / (double) BA.x();
		
		if(BA.y() != 0)
		{
			// everything normal...
			x2 = (double) PA.y() / (double) BA.y();
			if(x1 == x2 && paramBounds(x1))
			{
				paramVal = x1;
				return true;
			}
		}
		// BA.y == 0 but BA.x != 0
		else
		{
			if(PA.y() != 0) return false;
			if(paramBounds(x1))
			{
				paramVal = x1;
				return true;
			}
		}

	}
	// BA.x == 0
	else
	{
		if(PA.x() != 0) return false;
		else
		{
			if(BA.y() != 0)
			{
				x2 = (double) PA.y() / (double) BA.y();
				if(paramBounds(x2))
				{
					paramVal = x2;
					return true;
				}
			}
			// BA.y == 0 AND BA.x == 0
			// its a point!
			else
			{
				if(_A == point)
				{
					paramVal = 0;
					return true;
				}
			}
		}
	}
	return false;
}

inline Vertex Line::getNormal() const
{
	// always returns right-hand normal
	Vertex dir = getVector();
	return Vertex(dir.y(),-dir.x());
}

Vertex Line::getPoint(double param) const
{
	return _A + _B*param - _A*param;
}
inline Vertex Line::getVector() const
{
	return _B-_A;
}

double Line::getLength() const
{
	// Pythagorean theorem
	Vertex v(getVector());
	return sqrt((double) (v.x()*v.x()+v.y()*v.y()));
}

double Line::getAngle(const Line& other) const
{
	// returns the right-hand angle
	// can range from -180-180°

	Vertex P = getVector();
	Vertex Q = other.getVector();

	double a = atan2((double)P.y(), (double)P.x());
	double b = atan2((double)Q.y(), (double)Q.x());

	a = a-b;

	if(a > M_PI) a -= M_PI*2;
	if(a < -M_PI) a += M_PI*2;

	return a;
}

int Line::getDirection(const Vertex& point) const
{
	// with normal form
	// right of line -> 1
	// left of line -> -1
	// on line -> 0
	int NF = getNormal()*(point-_A);
	if(NF > 0) return 1;
	if(NF < 0) return -1;
	return 0;
}

Line::Line_type Line::getType() const
{
	return type;
}

void Line::setType(Line_type t)
{
	type = t;
}