#include "PolygonClipper.h"

#include <vector>
#include <cassert>
#define _USE_MATH_DEFINES
#include <cmath>

#include "Intersection.h"
#include "ringvec.h"

PolygonClipper::PolygonClipper(const Polygon &s, const Polygon &c)
: id(0),subject_polygon(s),clip_polygon(c),has_self_intersections_in_a(false),has_self_intersections_in_b(false),dont_delete_polygonlists(false)
{
	// check bounding boxes first
	if(!subject_polygon.insideBoundingBox(clip_polygon))
	{
		outside_bounding_box = true;
		return;
	}
	outside_bounding_box = false;

	// assert validity. This is only done in debug because it is
	// quite expensive to check all the constrictions
	#ifndef NDEBUG

	assert(subject_polygon.valid());
	assert(clip_polygon.valid());

	#endif

	// fill up augmented polygons
	for(std::vector<Vertex>::const_iterator it = subject_polygon.vertices().begin(); it != subject_polygon.vertices().end(); ++it)
		polygon_a.push_back(PolyVertex(*it,id++));

	for(std::vector<Vertex>::const_iterator it = clip_polygon.vertices().begin(); it != clip_polygon.vertices().end(); ++it)
		polygon_b.push_back(PolyVertex(*it,id++));

	has_intersections = calculateIntersections();

	// look if there are any self-intersections
	if(!self_intersections_a.empty())
	{
		splitUpAtSelfIntersections(polygon_a,polygons_of_a,self_intersections_a,false);
		eliminateHolesSimple(polygons_of_a);
		has_self_intersections_in_a = true;
	}
	if(!self_intersections_b.empty())
	{
		splitUpAtSelfIntersections(polygon_b,polygons_of_b,self_intersections_b,true);
		eliminateHolesSimple(polygons_of_b);
		has_self_intersections_in_b = true;
	}

}

PolygonClipper::~PolygonClipper()
{
	if(dont_delete_polygonlists) return;
	clearPolygonList(polygons_of_a);
	clearPolygonList(polygons_of_b);
}

void PolygonClipper::clearPolygonList(std::list<Polygon*> &polygons) const
{
	for(std::list<Polygon*>::iterator it = polygons.begin(); it != polygons.end(); ++it)
		delete *it;
	polygons.clear();
}

std::list<Polygon*> PolygonClipper::getSubjectPolygons() const
{
	std::list<Polygon*> polys;
	if(polygon_a.empty())
		for(std::list<Polygon*>::const_iterator it = polygons_of_a.begin(); it != polygons_of_a.end(); ++it)
			polys.push_back(new Polygon(**it));
	else
		polys.push_back(createPolygon(polygon_a));
	return polys;
}

std::list<Polygon*> PolygonClipper::getClipPolygons() const
{
	std::list<Polygon*> polys;
	if(polygon_b.empty())
		for(std::list<Polygon*>::const_iterator it = polygons_of_b.begin(); it != polygons_of_b.end(); ++it)
			polys.push_back(new Polygon(**it));
	else
		polys.push_back(createPolygon(polygon_b));
	return polys;
}

/* +++++++++++++++++++++++++++++++++ Main ++++++++++++++++++++++++++++++++++ */

Clip_result PolygonClipper::clip(Clipping_operation op, std::list<Polygon *> &result)
{
	// check bounding boxes first
	if(outside_bounding_box)
		return standardClipResult(op);

	Poly_tag tag = None;

	// if there are no intersections: tag all the same or cancel...
	if(!has_intersections)
	{
		Poly_tag tag_for_a = None;
		Poly_tag tag_for_b = None;
		// first determine their orientation to each other:
		// one point of other polygon is inside this polygon
		if(subject_polygon.contains(clip_polygon.vertices().front()))
		{
			tag_for_a = Outside;
			tag_for_b = Inside;
		}
		// one point of this polygon is inside other polygon
		else if(clip_polygon.contains(subject_polygon.vertices().front()))
		{
			tag_for_a = Inside;
			tag_for_b = Outside;
		}
		// if there are no intersections and any point of the polygon b and a is
		// not inside the other, it means the polygons do not actually intersect.
		// We can return standard results
		else
		{
			return standardClipResult(op);
		}
		// Otherwise - depending on the clipping operation - we can a shortcut
		// for some of the clipping operations here and return unmodified polygons
		Clip_result enclosed = enclosedClipResult(op,tag_for_a);
		if(enclosed != Default)
			return enclosed;

		// otherwise, tag all the edges the same and skip the
		// regular tagging
		for(polyvertex_it i = polygon_a.begin(); i != polygon_a.end(); ++i)
			i->tag = tag_for_a;
		for(polyvertex_it i = polygon_b.begin(); i != polygon_b.end(); ++i)
			i->tag = tag_for_b;

		tag = tag_for_a;
	}
	// there are several subject and clip polygons now... :-(
	else if(has_self_intersections_in_a || has_self_intersections_in_b)
	{
		return handleMultipleClippingOperations(op,result);
	}
	else
	{
		// tag the polygons
		tag = tagPolygons();

		// shortcuts:
		// Polygon A does only consist of shared edges (A clip A)
		if(tag == Shared)
		{
			// nothing
			if(op == Difference || op == Exclusion)
				return Nothing;
			// subject polygon
			if(op == Union || op == Intersect)
				return SubjectPolygon;
		}
		// Polygon A does only consist of outside edges (polygons
		// max. only touch each other at intersection points)
		else if(tag == Outside)
		{
			return standardClipResult(op);
		}
		// Polygon A does consist only of outside and reverse-shared
		// edges (polygons only touch each other)
		else if(tag == ReverseShared)
		{
			if(op == Difference)
				return SubjectPolygon;
			if(op == Intersect)
				return Nothing;
		}
		// Polygon A does consist only of inside and shared edges
		// (polygons only touch each other but one is contained in
		// the other)
		else if(tag == Inside)
		{
			Clip_result enclosed = enclosedClipResult(op,tag);
			if(enclosed != Default)
				return enclosed;
		}
	}

	// then, put the polygons together...
	std::list<Polygon *> clipResult;
	constructPolygons(op,clipResult);

	// if the operation is exclusion or union and the two polygons do not
	// actually intersect, BothPolygons should be returned if the result
	// contains a hole
	if(has_intersections)
	{
		if(tag == ReverseShared || tag == Outside)
			if(op == Exclusion || op == Union)
			{
				for(std::list<Polygon *>::iterator it = clipResult.begin(); it != clipResult.end(); ++it)
					if((*it)->size() <= 0)
						return BothPolygons;
			}
	}


	// for every hole: split polygons...
	eliminateHoles(clipResult);

	// at last, put result in the result list
	result.splice(result.end(),clipResult);

	return Default;
}

Clip_result PolygonClipper::enclosedClipResult(Clipping_operation op, Poly_tag tag_for_a) const
{
		if(op == Difference)
		{
			// Difference: if A is smaller, return nothing
			if(tag_for_a == Inside)
				return Nothing;
		}
		else if(op == Union)
		{
			// Union: return only the larger one
			if(tag_for_a == Inside) return ClipPolygon;
			else return SubjectPolygon;
		}
		else if(op == Intersect)
		{
			// Intersection: return only the smaller one
			if(tag_for_a == Outside) return ClipPolygon;
			else return SubjectPolygon;
		}
		return Default;
}

Clip_result PolygonClipper::standardClipResult(Clipping_operation op) const
{
	// these are the standard clip results for the different operations if the
	// two polygons do not intersect:
	if(op == Difference)
		return SubjectPolygon;
	if(op == Union)
		return BothPolygons;
	if(op == Exclusion)
		return BothPolygons;
	if(op == Intersect)
		return Nothing;
	return Default;
}


/* +++++++++++++++++++++++++ Step 1: Intersections +++++++++++++++++++++++++ */

bool PolygonClipper::calculateIntersections()
{
	/*
	Step 1: calculate intersections and insert them in the tagged
	polygons. Since we need to check intersections of every line of one
	polygon with every line of the other polygon, this is expensive but
	there is no other way to do that
	*/	
	bool hasIntersections = false;

	// for every line of Polygon A
	for(polyvertex_it i = polygon_a.begin(); i != polygon_a.end(); ++i)
	{
		// for every line of Polygon B
		for(polyvertex_it j = polygon_b.begin(); j != polygon_b.end(); ++j)
		{
			if(handleIntersection(i,j))
				hasIntersections = true;
		}
	}
	return hasIntersections;
}

// this method can be used instead of the above if it is known that the two
// polygons only intersect each other at their endpoints AND are otherwise
// valid polygons
bool PolygonClipper::calculateIntersectionsSimple()
{
	bool hasIntersections = false;

	// for every vertex of Polygon A
	for(polyvertex_it i = polygon_a.begin(); i != polygon_a.end(); ++i)
	{
		Vertex p(i->data);
		// for every vertex of Polygon B
		for(polyvertex_it j = polygon_b.begin(); j != polygon_b.end(); ++j)
		{
			Vertex q(j->data);
			if(p == q)
			{
				// link from PolyVertexLink to iterators at intersection
				intersections[p].positionsInA.insert(i);
				intersections[p].positionsInB.insert(j);
				// link from PolyVertices to which the iterators link
				// back to the PolyVertexLink
				i->link = &intersections[p];
				j->link = &intersections[p];
				hasIntersections = true;
			}
		}
	}
	return hasIntersections;
}

bool PolygonClipper::handleIntersection(polyvertex_it &i, polyvertex_it &j)
{

	Vertex P = decreaseIt(j,polygon_b)->data;
	Vertex Q = j->data;

	Vertex A = decreaseIt(i,polygon_a)->data;
	Vertex B = i->data;

	Line edge_in_a(A,B,Line::Segment);
	Line edge_in_b(P,Q,Line::Segment);

	Intersection intersect;

	// not found intersection
	if(!edge_in_a.intersects(edge_in_b,intersect)) return false;

	// for points: insert if not already there
	if(!intersect.isLineIntersection())
	{
		Vertex p = *(intersect.point);
		polyvertex_it posA, posB;

		handlePointIntersection(p,edge_in_a,i,polygon_a,posA);
		handlePointIntersection(p,edge_in_b,j,polygon_b,posB);

		// if the point of the intersection has been created at it's position without
		// rounding, memorize the intersection
		if(intersect.exact)
		{
			assert(posA->data == posB->data);
			Vertex key(posA->data);
			// link from PolyVertexLink to iterators at intersection
			intersections[key].positionsInA.insert(posA);
			intersections[key].positionsInB.insert(posB);
			// link from PolyVertices to which the iterators link
			// back to the PolyVertexLink
			posA->link = &intersections[key];
			posB->link = &intersections[key];
		}
		// otherwise, we have to first check for more intersections because the shape
		// of the polygon just changed.
		else
		{
			checkForNewIntersections(posA, posB);
		}
	}
	// for lines 
	else
	{
		Line l = *(intersect.line);

		polyvertex_it posA1;
		polyvertex_it posA2;
		polyvertex_it posB1;
		polyvertex_it posB2;

		handleLineIntersection(l, edge_in_a, i, polygon_a, posA1, posA2);
		handleLineIntersection(l, edge_in_b, j, polygon_b, posB1, posB2);

		// (if we have a line intersection, the two points are always created without rounding
		// imprecision. Thats why we memorize the intersection directly and don't check recursively)

		// insert links from PolyVertexLink to (iterators to) PolyVertices and back.
		intersections[posA1->data].positionsInA.insert(posA1);
		intersections[posA2->data].positionsInA.insert(posA2);
		intersections[posB1->data].positionsInB.insert(posB1);
		intersections[posB2->data].positionsInB.insert(posB2);
		posA1->link = &intersections[posA1->data];
		posA2->link = &intersections[posA2->data];
		posB1->link = &intersections[posB1->data];
		posB2->link = &intersections[posB2->data];
	}

	return true;
}

void PolygonClipper::handlePointIntersection(	const Vertex &intersectPoint,	const Line &line,
																							polyvertex_it &insert, std::list<PolyVertex> &poly,
																							polyvertex_it &pos) 
{
	// first point of the line segment counts as the last point of the
	// last line segment
	if(intersectPoint == line.A())
		pos = decreaseIt(insert,poly);
	else if(intersectPoint == line.B())
		pos = insert;
	// if intersection on line, insert it
	else
		insert = pos = poly.insert(insert,PolyVertex(intersectPoint,id++));
}

void PolygonClipper::handleLineIntersection(	const Line &intersectLine, const Line &line,
																							polyvertex_it &insert, std::list<PolyVertex> &poly,
																							polyvertex_it &pos1, polyvertex_it &pos2) 
{
	// make intersection line go into the same direction as this line
	//(makes if-queries easier)
	bool oppositeDirection = line.getVector()*intersectLine.getVector() < 0;

	Vertex A = line.A();
	Vertex B = line.B();

	Vertex p = oppositeDirection ? intersectLine.B() : intersectLine.A();
	Vertex q = oppositeDirection ? intersectLine.A() : intersectLine.B();

	polyvertex_it i = insert;

	// insert both points
	if(p != A && q != B)
	{
		i = pos1 = poly.insert(insert,PolyVertex(p,id++));
		pos2 = poly.insert(insert,PolyVertex(q,id++));
	}
	// both points already exist
	else if(p == A && q == B)
	{
		pos1 = decreaseIt(insert,poly);
		pos2 = insert;
	}
	// one of the points must be inserted
	else if(p == A)
	{
		pos1 = decreaseIt(insert,poly);
		i = pos2 = poly.insert(insert,PolyVertex(q,id++));
	}
	else if(q == B)
	{
		i = pos1 = poly.insert(insert,PolyVertex(p,id++));
		pos2 = insert;
	}

	// The current polyvertex_it must be set to the first inserted vertex
	// so that no edges are forgotten during the intersection checks.
	// If none have been inserted, the current polyvertex_it it not changed.
	insert = i;
}



bool PolygonClipper::handleSelfIntersection(	polyvertex_it &it, polyvertex_it &jt, bool forB	)
{
	if(it == jt) return false;

	std::list<PolyVertex> &poly = forB ? polygon_b : polygon_a;

	polyvertex_it it_before = decreaseIt(it,poly);
	polyvertex_it jt_before = decreaseIt(jt,poly);

	Line line1(it_before->data,it->data,Line::Segment);
	Line line2(jt_before->data,jt->data,Line::Segment);

	Intersection intersect;
	if(!line1.intersects(line2,intersect)) return false;

	// every line intersects with it's successor and the one behind.
	// we can ignore that
	if(!intersect.isLineIntersection())
	{
		if(increaseIt(it,poly) == jt && intersect.thisParam == 1.0)
			return false;
		if(it_before == jt && intersect.thisParam == 0.0)
			return false;
	}

	std::map<Vertex,PolyVertexLink,VertexComp> &self_intersections = forB ? self_intersections_b : self_intersections_a;

	// line intersection
	if(intersect.isLineIntersection())
	{
		polyvertex_it line1_pos1, line1_pos2, line2_pos1, line2_pos2;

		handleLineIntersection(*intersect.line,line1,it,poly,line1_pos1, line1_pos2);
		handleLineIntersection(*intersect.line,line2,jt,poly,line2_pos1, line2_pos2);

		// insert links from PolyVertexLink to (iterators to) PolyVertices and back.
		// only insert if these are really two different PolyVertices
		if(line1_pos1 != line2_pos2)
		{
			self_intersections[line1_pos1->data].positionsInA.insert(line1_pos1);
			self_intersections[line2_pos2->data].positionsInA.insert(line2_pos2);
			line1_pos1->selflink = &self_intersections[line1_pos1->data];
			line2_pos2->selflink = &self_intersections[line2_pos2->data];
		}
		// only insert if these are really two different PolyVertices
		if(line1_pos2 != line2_pos1)
		{
			self_intersections[line1_pos2->data].positionsInA.insert(line1_pos2);
			self_intersections[line2_pos1->data].positionsInA.insert(line2_pos1);
			line1_pos2->selflink = &self_intersections[line1_pos2->data];
			line2_pos1->selflink = &self_intersections[line2_pos1->data];
		}
	}
	// point intersection
	else
	{
		polyvertex_it line1_pos, line2_pos;
		
		handlePointIntersection(*intersect.point, line1, it, poly, line1_pos);
		handlePointIntersection(*intersect.point, line2, jt, poly, line2_pos);

		assert(line1_pos->data == line2_pos->data);
		Vertex key(line1_pos->data);
		// link from PolyVertexLink to iterators at intersection
		self_intersections[key].positionsInA.insert(line1_pos);
		self_intersections[key].positionsInA.insert(line2_pos);
		// link from PolyVertices to which the iterators link
		// back to the PolyVertexLink
		line1_pos->selflink = &self_intersections[key];
		line2_pos->selflink = &self_intersections[key];

		// if the self-intersection is not exact (=shape changed)
		// we need to check for new intersections recursively
		// (however, in this case only the shape of one polygon
		// has changed, so we only need to check that)
		if(!intersect.exact)
		{
			if(!forB)
			{
				polyvertex_it nada =  polygon_b.end();
				checkForNewIntersections(	line1_pos, nada);
				checkForNewIntersections(	line2_pos, nada);
			}
			else
			{
				polyvertex_it nada =  polygon_a.end();
				checkForNewIntersections(	nada, line1_pos );
				checkForNewIntersections(	nada, line2_pos );
			}
		}
	}
	return true;
}



void PolygonClipper::checkForNewIntersections(	polyvertex_it &posA, polyvertex_it &posB )
{
	if(posA != polygon_a.end())
	{
		polyvertex_it n_posA = increaseIt(posA,polygon_a);

		// check for self-intersections in polygon A
		for(polyvertex_it c = polygon_a.begin(); c != polygon_a.end(); ++c)
		{
			handleSelfIntersection(posA, c, false);
			handleSelfIntersection(n_posA, c, false);
		}
		// check for intersections between the 2 new edges in polygon A with any
		// edges in polygon B
		for(polyvertex_it c = polygon_b.begin(); c != polygon_b.end(); ++c)
		{
			handleIntersection(posA,c);
			handleIntersection(n_posA,c);
		}
	}
	if(posB != polygon_b.end())
	{
		polyvertex_it n_posB = increaseIt(posB,polygon_b);

		// check for self-intersections in polygon B
		for(polyvertex_it c = polygon_b.begin(); c != polygon_b.end(); ++c)
		{
			handleSelfIntersection(posB, c, true);
			handleSelfIntersection(n_posB, c, true);
		}
		// same with other
		for(polyvertex_it c = polygon_a.begin(); c != polygon_a.end(); ++c)
		{
			handleIntersection(c,posB);
			handleIntersection(c,n_posB);
		}
	}
}

void PolygonClipper::splitUpAtSelfIntersections(
																			std::list<PolyVertex> &poly, 
																			std::list<Polygon*> &polygons,
																			std::map<Vertex,PolyVertexLink,VertexComp> &self_intersections,
																			bool forB)
{
	std::list<polyvertex_it> polystack;
	std::list<std::list<PolyVertex>> polys;

	// start at one self intersection
	polyvertex_it start = *self_intersections.begin()->second.positionsInA.begin();
	polyvertex_it it = start;
	polystack.push_back(start);
	polyvertex_it last_before_start = poly.end();

	do
	{
		// goto marker to start of loop
		start_of_loop:

		// walk until next self-intersection has been found
		polyvertex_it next_it = increaseIt(it,poly);
		// remember after which intersection we encountered the start of the polygon
		if(next_it == poly.begin()) last_before_start = polystack.back();

		if(next_it->selflink)
		{
			// look if we got the start point already in the "stack"
			std::list<polyvertex_it>::reverse_iterator stack_it;
			bool same_point_exists = false;
			for(stack_it = polystack.rbegin(); stack_it != polystack.rend(); ++stack_it)
			{
				bool same_point = (next_it->data == (*stack_it)->data);
				same_point_exists |= same_point;
				// if we reached a self-intersection which is at the same point as one of those
				// we started with, finalize polygon 
				if(same_point)
				{
					polys.push_back(std::list<PolyVertex>());
					std::list<PolyVertex> &mypoly = polys.back();
					// ...from top of stack ... to ... current vertex
					polyvertex_it polygon_start = *stack_it;
					
					// remove link to self-intersection because this one will be solved now
					// and pop everything from the "stack" that came after the one we used
					while(polystack.back() != polygon_start)
					{
						clearSelfIntersectionLinks(polystack.back(),self_intersections);
						polystack.pop_back();
					}	
					int links_left = clearSelfIntersectionLinks(polygon_start,self_intersections);
					polystack.pop_back();
					
					// finish up polygon (the next vertex is the vertex we started with,
					// this can only mean that we have gone round completely)
					if(polygon_start == next_it)
					{
						mypoly.splice(mypoly.begin(),poly);
					}
					// otherwise splice normally
					else if(last_before_start != polygon_start)
					{
						mypoly.splice(mypoly.begin(),poly,polygon_start,next_it);
					}
					// splice over the end
					else
					{
						mypoly.splice(mypoly.begin(),poly,polygon_start,poly.end());
						if(next_it != poly.begin())
							mypoly.splice(mypoly.end(),poly,poly.begin(),next_it);
						// we are exactly at the start now. This means the last intersection point
						// before the start is the point on the top of the stack (if any)
						if(!polystack.empty())
							last_before_start = polystack.back();
						else
							last_before_start = poly.end();
					}

					// our start vertex has been popped
					if(polystack.empty()) 
					{
						start = next_it;
						polystack.push_back(it = next_it);
						goto start_of_loop;
					}
					// another "pimple" will follow because there are selfintersections at this point left
					else if(links_left > 0)
					{
						polystack.push_back(it = next_it);
						goto start_of_loop;
					}
					break;
				}
			}
			// otherwise, there must be a "pimple" on that one. Push a new start vertex
			// to the stack
			if(!same_point_exists)
			{
				polystack.push_back(next_it);
			}
		}
		it = next_it;

	} while(it != start);

	// put remaining vertices into a polygon
	polys.push_back(std::list<PolyVertex>());
	std::list<PolyVertex> &restpoly = polys.back();
	restpoly.splice(restpoly.begin(),poly);

	assert(polys.size() > 1);
	assert(poly.empty());

	// put into list of Polygon*
	for(std::list<std::list<PolyVertex> >::iterator it = polys.begin(); it != polys.end(); ++it)
	{
		// filter out polygons with zero size (can only be: those with two vertices)
		if(it->size() > 2)
			polygons.push_back(createPolygon(*it));
	}
}

void PolygonClipper::eliminateHolesSimple(std::list<Polygon *> &polygons) const
{
	std::list<Polygon*> holes;
	std::list<Polygon*> result;
	for(std::list<Polygon*>::iterator it = polygons.begin(); it != polygons.end(); )
	{
		if((*it)->size() < 0)
		{
			Polygon *ccwHole = new Polygon(-(**it));
			holes.push_back(ccwHole);
			delete *it;
			it = polygons.erase(it);
		}
		else
			++it;
	}
	if(!holes.empty())
	{
		clipMultiple(Difference,polygons,holes,result);
		polygons.splice(polygons.begin(),result);
	}
}

int PolygonClipper::clearSelfIntersectionLinks(polyvertex_it &polygon_start,
																								std::map<Vertex,PolyVertexLink,VertexComp> &self_intersections)
{
		// first remove the given link
		PolyVertexLink *selfintersectionlink = polygon_start->selflink;
		selfintersectionlink->positionsInA.erase(polygon_start);
		polygon_start->selflink = NULL;

		// then check:
		// if there is only one point in the self-intersection left, it is no
		// self intersection anymore. The whole self-intersection can be removed
		int size = selfintersectionlink->positionsInA.size();
		if(size == 1)
		{
			polyvertex_it otherpoint(*selfintersectionlink->positionsInA.begin());
			otherpoint->selflink = NULL;
			self_intersections.erase(polygon_start->data);
			return 0;
		}
		return size;
}


/* ++++++++++++++++++++++++++++ Step 2: Tagging ++++++++++++++++++++++++++++ */

PolygonClipper::Poly_tag PolygonClipper::tagPolygons()
{
	/*
		Step 2: Tag polygons. For each line of both polygons we need to find out
		if they are inside, outside, shared or reverse shared line segments...
	*/

	bool aOnlyOutside = true, aOnlyShared = true,
	     aOnlyOutsideOrReverseShared = true, aOnlyInsideOrShared = true;

	// for all intersections between the two polygons...
	for(polyvertexlink_it it = intersections.begin(); it != intersections.end(); ++it)
	{

		PolyVertexLink &pvl(it->second);
		for(intersect_pos_it at = pvl.positionsInA.begin(); at != pvl.positionsInA.end(); ++at)
		{
			Poly_tag tag = tagForIntersection(*at,false);

			// if all edges of the subject polygon are tagged the same, certain
			// shortcuts can be made (see below)
			if(tag == None) continue;
			if(tag != Shared)
			{
				aOnlyShared = false;
				if(tag != Inside)
					aOnlyInsideOrShared = false;
			}
			if(tag != Outside)
			{
				aOnlyOutside = false;
				if(tag != ReverseShared)
					aOnlyOutsideOrReverseShared = false;
			}
		}
		for(intersect_pos_it bt = pvl.positionsInB.begin(); bt != pvl.positionsInB.end(); ++bt)
		{
			tagForIntersection(*bt,true);
		}
	}

#ifndef NDEBUG

	// all edges must be tagged at this point
	for(polyvertex_it it = polygon_a.begin(); it != polygon_a.end(); ++it)
		assert(it->tag != PolygonClipper::None);
	for(polyvertex_it jt = polygon_b.begin(); jt != polygon_b.end(); ++jt)
		assert(jt->tag != PolygonClipper::None);

#endif

	// shortcuts for certain tagggings of polygon A
	if(aOnlyShared) return Shared;
	if(aOnlyOutside) return Outside;
	if(aOnlyOutsideOrReverseShared) return ReverseShared;
	if(aOnlyInsideOrShared) return Inside;
	return None;
}

PolygonClipper::Poly_tag PolygonClipper::tagForIntersection(const polyvertex_it &p, bool forB)
{
	polyvertex_it nextP = forB ? increaseIt(p,polygon_b) : increaseIt(p,polygon_a);

	// skip if for this intersection, we already tagged the edges
	// (happens when at this point, we have multiple intersections with the other polygon)
	if (p->tag != None)
		return None;

	// find the two edges of the other polygon which are closest to the line which interests us now
	// normally there are just two edges but theoretically, there can be several for weakly simple
	// polygons
	Line thisLine(p->data,nextP->data);

	Line prevLine;
	Line nextLine;

	std::set<polyvertex_it,PolyVertexItComp> &verticesInOtherPolygon = forB ? p->link->positionsInA : p->link->positionsInB;
	std::list<PolyVertex> &poly = forB ? polygon_a : polygon_b;

	calculateClosestEdgesToLine(thisLine, prevLine, nextLine, verticesInOtherPolygon, poly, forB);

	// determine if the current line is inside or outside the other polygon by
	// looking at the two surrounding edges of the other polygons
	Poly_tag tag = getLineOrientation(prevLine,nextLine,thisLine);

	// now tag the polygon till the next intersection is found
	polyvertex_it pit = p;
	do {
		pit->tag = tag;
		// we increase the iterator with this function instead of ++ because
		// a polygon is a ring - if the end of the list of vertices has been
		// reached, he should just start at the front again.
		pit = forB ? increaseIt(pit,polygon_b) : increaseIt(pit,polygon_a);
	} while(!pit->link);

	return tag;
}

void PolygonClipper::calculateClosestEdgesToLine(const Line& line, Line &prevLine, Line &nextLine,
																					std::set<polyvertex_it,PolyVertexItComp> &verticesInOtherPolygon,
																					std::list<PolyVertex> &otherPolygon, bool forB) const
{
		double min_prev_angle = 10.0, max_prev_angle = -10.0;
		double min_next_angle = 10.0, max_next_angle = -10.0;

		Line min_prev, max_prev, min_next, max_next;

		// for all vertices in the other polygon which are referenced by the intersection point from A...
		for(intersect_pos_it jt = verticesInOtherPolygon.begin(); jt != verticesInOtherPolygon.end(); ++jt )
		{
			polyvertex_it other = (*jt);

			// get the previous and next vertex of the other polygon...
			polyvertex_it other_previous = decreaseIt(other,otherPolygon);
			polyvertex_it other_next = increaseIt(other,otherPolygon);

			Vertex A(other_previous->data);
			Vertex P(other->data);
			Vertex B(other_next->data);
			Line curr_prevLine(A,P);
			Line curr_nextLine(P,B);

			// and get their angles towards the line
			double curr_prev_angle = line.getAngle(curr_prevLine);
			double curr_next_angle = line.getAngle(curr_nextLine);
			if(curr_next_angle < 0) curr_next_angle += 2.0*M_PI;

			// determine previous line with minimum angle
			if(curr_prev_angle < min_prev_angle)
			{
				min_prev_angle = curr_prev_angle;
				min_prev = curr_prevLine;
			}
			// determine previous line with maximum angle
			if(curr_prev_angle > max_prev_angle)
			{
				max_prev_angle = curr_prev_angle;
				max_prev = curr_prevLine;
			}
			// determine next line with minimum angle
			if(curr_next_angle < min_next_angle)
			{
				min_next_angle = curr_next_angle;
				min_next = curr_nextLine;
			}
			// determine next line with maximum angle
			if(curr_next_angle > max_next_angle)
			{
				max_next_angle = curr_next_angle;
				max_next = curr_nextLine;
			}
		}
		// of the two sets of lines selected, select the set which encloses the
		// the smallest angle
		double max_prev_min_next_angle = getEnclosingAngle(max_prev_angle, min_next_angle);
		double min_prev_max_next_angle = getEnclosingAngle(min_prev_angle, max_next_angle);

		if(max_prev_min_next_angle < min_prev_max_next_angle)
		{
			prevLine = max_prev;
			nextLine = min_next;
		}
		else
		{
			prevLine = min_prev;
			nextLine = max_next;
		}
}

double PolygonClipper::getEnclosingAngle(double prev, double next) const
{
	if(prev + M_PI > next)
		return M_PI - prev + next;
	else
		return M_PI + prev - next + M_PI*2.0;
}

PolygonClipper::Poly_tag PolygonClipper::getLineOrientation(const Line &prevLine, const Line &nextLine,
																														const Line &line) const
{
		// its on the line itself?
		// same direction as the other lines
		if(line.B() == nextLine.B()) return Shared;
		// opposite direction as the other lines
		if(line.B() == prevLine.A()) return ReverseShared;

		bool line1inside = prevLine.getDirection(line.B()) <= 0;
		bool line2inside = nextLine.getDirection(line.B()) <= 0;
		bool linesconcave =  prevLine.getAngle(nextLine) > 0;

		Poly_tag result = Outside;

		// if the other lines are concave, the line must be on the left side of
		// either of the two
		if(linesconcave)
		{
			if(line1inside || line2inside)
				result = Inside;
		}
		// otherwise must be on the left side of both
		else
		{
			if(line1inside && line2inside)
				result = Inside;
		}
		return result;
}


/* +++++++++++++++++++++ Step 3: Construct polygon(s) ++++++++++++++++++++++ */

void PolygonClipper::constructPolygons(Clipping_operation op, std::list<Polygon*> &clipResult)
{
	// all vertex_links are iterated
	for(polyvertexlink_it it = intersections.begin(); it != intersections.end(); ++it)
	{
		PolyVertexLink *link = &(it->second);

		// if the vertex_link has already been visited in below loop: to next intersection
		if(link->visited) continue;

		// the polygon(s) are constructed in this list, one vertex after one
		// another is added during the process.
		std::list<std::list<Vertex>> construct;

		std::list<polyvertex_it> polystack;

		// in one cycle of this loop, we select one path and walk along it until
		// we reach the next intersection. We do that until we reach an
		// intersection which has already been visited
		do {

			// select the egible paths we can walk. Whats egible and what's not 
			// depends on the clipping operation we chose.
			std::list<PolyPathInfo> availPaths = getAllAvailablePaths(link,op);
			// no available paths? skip
			if(availPaths.empty())
			{
				link->visited = true;
				continue;
			}
			// choose the path which goes - relative to the last line added to
			// the polygon in construction - into the leftmost direction
			// If this will be the first line of the polygon, choose any path
			std::list<PolyPathInfo>::iterator best = availPaths.begin();

			if(construct.size() > 0 && availPaths.size() > 1)
			{
				double bestAngle = M_PI;
				// last line: A = last vertex in construct; B = any walker data of the available paths
				Line last(*(construct.rbegin()->rbegin()), best->walker->data);

				// this loop should iterate over 1-4 paths for simple polygons (and any number for weakly simple polygons)
				for(std::list<PolyPathInfo>::iterator tri = availPaths.begin(); tri != availPaths.end(); ++tri)
				{
					// walk along this path to get the line
					Vertex A(tri->walker->data), B;
					if(!tri->reverse)
						B = increaseIt(tri->walker,tri->path)->data;
					else
						B = decreaseIt(tri->walker,tri->path)->data;
					Line line(A,B);
					double angle = last.getAngle(line);

					if(angle <= bestAngle)
					{
						bestAngle = angle;
						best = tri;
					}
				}
			}
			// left-most path selected
			PolyPathInfo *c = &(*best);
			polystack.push_back(c->walker);

			// mark selected path as visited: tag the edge as None so that it is not
			// found again later and if it was the only possible path, tag the
			// vertex_link as visited
			if(c->reverse)
				decreaseIt(c->walker,c->path)->tag = None;
			else
				c->walker->tag = None;
				
			if(availPaths.size() == 1)
				link->visited = true;

			// walk along the path until next intersection and put all vertices that
			// are on the way into construct
			construct.push_back(std::list<Vertex>());
			do
			{
				construct.rbegin()->push_back(c->walker->data);
				if(c->reverse)
					c->walker = decreaseIt(c->walker,c->path);
				else
					c->walker = increaseIt(c->walker,c->path);
			}
			while(!c->walker->link);

			// new intersection reached
			link = c->walker->link;
			
			// if we reached a starting point of the polygon we are constructing,
			// finalize polygon
			std::list<polyvertex_it>::iterator sit;
			std::list<std::list<Vertex>>::iterator cit;
			for(sit = polystack.begin(), cit = construct.begin();	sit != polystack.end(); ++sit, ++cit)
			{
				if(c->walker->data == (*sit)->data)
				{
					// splice together part of the construct that should end up as a new polygon
					std::list<Vertex> final;
					for(std::list<std::list<Vertex>>::iterator splice_it = cit; splice_it != construct.end(); ++splice_it)
					{
						final.splice(final.end(),*splice_it);
						polystack.pop_back();
					}
					assert(cit != construct.end());
					construct.erase(cit,construct.end());

					Polygon *poly = new Polygon(final);
					clipResult.push_back(poly);
					break;
				}
			}

		} while(!link->visited);

		assert(construct.empty());
		assert(polystack.empty());
	}
	// no intersections at all. At this point, this can only occur
	// for difference or exclusion. The other operations return earlier
	if(intersections.empty())
	{
		assert(op == Difference || op == Exclusion);
		// construct polygons: Inside = reverse
		//                     Outside = normal
		Polygon *first;
		Polygon *second;
		// polygon_a == subject polygon
		if(polygon_a.front().tag == Outside)
			first = new Polygon(subject_polygon);
		else
			first = new Polygon(-subject_polygon);

		// polygon_b == clip polygon
		if(polygon_b.front().tag == Outside)
			second = new Polygon(clip_polygon);
		else
			second = new Polygon(-clip_polygon);

		clipResult.push_back(first);
		clipResult.push_back(second);
	}
}


std::list<PolygonClipper::PolyPathInfo> PolygonClipper::getAllAvailablePaths(	PolyVertexLink* link,
																																							Clipping_operation op)
{
	std::list<PolyPathInfo> availPaths;

	// first look at all paths in A (normally, it is just one but for weakly simple
	// polygons, theoretically there can be any number)
	for(intersect_pos_it at = link->positionsInA.begin(); at != link->positionsInA.end(); ++at)
	{
		polyvertex_it walker = *at;
		Poly_tag next = walker->tag;
		Poly_tag prev = decreaseIt(walker,polygon_a)->tag;

		if (op == Difference)
		{
			if(next == Outside || next == ReverseShared)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
		}
		else if(op == Union)
		{
			if(next == Outside || next == Shared)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
		}
		else if(op == Intersect)
		{
			if(next == Inside || next == Shared)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
		}
		else if(op == Exclusion)
		{
			if(next == Outside)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,false));
			if(prev == Inside)
				availPaths.push_back(PolyPathInfo(polygon_a,walker,true));
		}
	}
	// then, for all in B
	for(intersect_pos_it bt = link->positionsInB.begin(); bt != link->positionsInB.end(); ++bt)
	{
		polyvertex_it walker = *bt;
		Poly_tag next = walker->tag;
		Poly_tag prev = decreaseIt(walker,polygon_b)->tag;

		if (op == Difference)
		{
			if(prev == Inside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,true));
		}
		else if(op == Union)
		{
			if(next == Outside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,false));
		}
		else if(op == Intersect)
		{
			if(next == Inside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,false));
		}
		else if(op == Exclusion)
		{
			if(next == Outside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,false));
			if(prev == Inside)
				availPaths.push_back(PolyPathInfo(polygon_b,walker,true));
		}
	}
	return availPaths;
}


/* +++++++++++++++++++++++ Step 4: Eliminate holes +++++++++++++++++++++++++ */



void PolygonClipper::eliminateHoles(std::list<Polygon *> &clipResult) const
{
	// 1. determine which polygons are the holes and put them in a separate list (ccw defined)
	std::list<Polygon *> holes;
	for(std::list<Polygon *>::iterator it = clipResult.begin(); it != clipResult.end(); )
	{
		Polygon *resultPolygon = *it;
		if(resultPolygon->size() < 0)
		{
			// NOTE: a new polygon wouldn't have to be created here if our clipping algorithm
			// worked with holes.
			Polygon *ccwHole = new Polygon(-(*resultPolygon));
			holes.push_back(ccwHole);
			it = clipResult.erase(it);
			delete resultPolygon;
		}
		else
			it++;
	}
	if(holes.empty()) return;

	// 2. for every hole there is, split all polygons
	for(std::list<Polygon *>::iterator holesit = holes.begin(); holesit != holes.end(); ++holesit)
	{
		Polygon *hole = *holesit;

		// determine split line (currently: just first edge of hole)
		std::vector<Vertex>::const_iterator v = hole->vertices().begin();
		Vertex first = *v;
		Vertex second = *(++v);
		Line splitLine(first,second,Line::Normal);

		std::list<Polygon *> splitResult;
		for(std::list<Polygon *>::iterator polyit = clipResult.begin(); polyit != clipResult.end();)
		{
			Polygon *polygon = *polyit;
			// polygon is split -> result is put into splitResult 
			if(polygon->split(splitLine,splitResult) != SubjectPolygon)
			{
				// delete old polygon
				polyit = clipResult.erase(polyit);
				delete polygon;
				if(clipResult.empty()) break;
			}
			else ++polyit;
		}
		// put the results into clipResult
		clipResult.splice(clipResult.end(),splitResult);
	}

	// 3. subtract every hole from all of the split up polygons
	std::list<Polygon *> subtractHoleResult;
	clipMultiple(Difference,clipResult,holes,subtractHoleResult);
	clipResult.splice(clipResult.end(),subtractHoleResult);

	#ifndef NDEBUG

	// there must be no holes in the output
	for(std::list<Polygon *>::const_iterator it = clipResult.begin(); it != clipResult.end(); ++it)
		assert((*it)->size() > 0);

	#endif
}

/* ++++++++++++++++++ Special: handle multiple intersectins ++++++++++++++++ */

Clip_result PolygonClipper::handleMultipleClippingOperations(Clipping_operation op, std::list<Polygon *> &result)
{
	// first put the remaining polygon_a/polygon_b into the list
	if(!polygon_a.empty())
	{
		assert(polygon_b.empty());
		polygons_of_a.push_back(createPolygon(polygon_a));
		polygon_a.clear();
	}
	if(!polygon_b.empty())
	{
		assert(polygon_a.empty());
		polygons_of_b.push_back(createPolygon(polygon_b));
		polygon_b.clear();
	}
	// now, do the operation:
	std::list<Polygon *> multipleClipResult;
	Clip_result cr = clipMultiple(op, polygons_of_a, polygons_of_b, multipleClipResult);
	// only put the clipping result into the result list if something was clipped
	// because the clip(...)-method should guarantee that the result-list is not altered
	// if cr != Default
	if(cr == Default) result.splice(result.end(),multipleClipResult);
	return cr;
}

Clip_result PolygonClipper::clipMultiple(Clipping_operation op,
																				 std::list<Polygon *> &polys_a,
																				 std::list<Polygon *> &polys_b,
																				 std::list<Polygon *> &result,
																				 bool simpleIntersection) const
{
	// what if no polygons are left? (cause there were invalid polygons)
	bool a_empty = polys_a.empty();
	bool b_empty = polys_b.empty();

	if(b_empty)
	{
		if(op == Intersect || a_empty)
		{
			clearPolygonList(polys_a);
			return Nothing;
		}
		else
		{
			result.splice(result.begin(),polys_a);
			return SubjectPolygon;
		}
	}
	if(a_empty)
	{
		
		if(op == Intersect || op == Difference || b_empty)
		{
			clearPolygonList(polys_b);
			return Nothing;
		}
		else
		{
			result.splice(result.begin(),polys_b);
			return ClipPolygon;
		}
	}

	assert(!polys_a.empty());
	assert(!polys_b.empty());

	Clip_result total_result;

	// for difference, subtract everything in B from each in A
	if(op == Difference)
	{
		total_result = SubjectPolygon;

		for(std::list<Polygon *>::iterator bit = polys_b.begin(); bit != polys_b.end(); ++bit)
		{
			for(std::list<Polygon *>::iterator ait = polys_a.begin(); ait != polys_a.end(); )
			{
				std::list<Polygon *> clipResult;
				Clip_result cr = (*ait)->clip(**bit,Difference,clipResult);
				// if something has been clipped, remove the original subject polygon
				// and replace it with the clip result in the list
				if(cr == Default || cr == Nothing)
				{
					delete *ait;
					ait = polys_a.erase(ait);
					total_result = Default;
					polys_a.splice(polys_a.end(),clipResult);
				}
				else
					++ait;
			}
		}
		// clear polygons in polys_b
		for(std::list<Polygon *>::iterator bit = polys_b.begin(); bit != polys_b.end(); ++bit)
		{
			delete *bit;
		}
		polys_b.clear();
		if(polys_a.empty())
			total_result = Nothing;
		result.splice(result.end(),polys_a);
	}

	// for union, try to merge as much as possible
	else if(op == Union)
	{
		total_result = BothPolygons;
		std::list<Polygon *> union_result;

		// first, put all in union_result
		union_result.splice(union_result.begin(),polys_b);
		std::list<Polygon *>::iterator it_start2 = union_result.begin();
		std::list<Polygon *>::iterator it_end = it_start2;
		union_result.splice(union_result.begin(),polys_a);

		bool clippedSomething = false;

		// Then for every polygon in union_result, try to merge it with any other polygon in union_result
		// Since we know that all polygons originally in polys_a are disjunct, we don't need to
		// try to merge those in the first run. Then, for any subsequent run, try to merge the newly added
		// polygons with the previous ones until no new clipping is done
		bool separate = true;
		do
		{
			clippedSomething = false;
			std::list<Polygon *> interimResult;

			for(std::list<Polygon *>::iterator it = union_result.begin(); it != it_end; )
			{
				bool gotonext = true;

				std::list<Polygon *>::iterator it2(it);
				++it2;
				if(separate)
					it2 = it_start2;

				for(; it2 != union_result.end(); )
				{
					bool gotonext2 = true;
					// if a union returns also 2 (or more) polygons, it doesn't make sense to apply it.
					// This happens if polygons are merged but create holes. Worse, it can lead to 
					// this loop never-ending because it switches between two stati all the time
					std::list<Polygon *> clipResult;
					Clip_result cr = (*it)->clip(**it2,Union,clipResult);
					// delete polygons if something has been clipped and add the new ones
					if(cr == Default && clipResult.size() < 2)
					{
						total_result = Default;
						clippedSomething = true;
						// remove the original polygons
						delete *it;
						delete *it2;
						// update iterators also
						bool update_start2 = (separate && it2 == it_start2);
						bool update_end = (it2 == it_end);
						std::list<Polygon *>::iterator tmp = union_result.erase(it2); 
						if(update_start2)
							it_start2 = tmp;
						if(update_end)
							it_end = tmp;
						it = union_result.erase(it);
						
						interimResult.splice(interimResult.end(),clipResult);
						gotonext = false;
						break;
					}
					// if only clip polygon is left -> put clip polygon in union_result
					// and delete the subject polygon
					else if(cr == ClipPolygon)
					{
						if(total_result == BothPolygons || total_result == ClipPolygon)
							total_result = ClipPolygon;
						else
							total_result = Default;
			
						delete *it;
						it = union_result.erase(it);
						gotonext = false;
						break;
					}
					// if only subject polygon is left, simply delete the clip polygon
					// and continue
					else if(cr == SubjectPolygon)
					{
						if(total_result == BothPolygons || total_result == SubjectPolygon)
							total_result = SubjectPolygon;
						else
							total_result = Default;
						
						delete *it2;
						// this is because if we deleted the iterator to it_end, we need to
						// update  it to the next iterator
						bool update_end = (it2 == it_end);
						it2 = union_result.erase(it2);
						if(update_end) it_end = it2;

						gotonext2 = false;
					}
					if(gotonext2) it2++;
				}
				if(gotonext) it++;
			}
			it_end = union_result.begin();
			union_result.splice(union_result.begin(),interimResult);
			separate = false;

		} while(clippedSomething);

		result.splice(result.end(),union_result);
	}
	// for intersect, intersect each in A with each in B
	else if(op == Intersect)
	{
		total_result = Nothing;

		std::list<std::list<Polygon *>::iterator> noclear_a;
		std::list<std::list<Polygon *>::iterator> noclear_b;

		for(std::list<Polygon *>::iterator ait = polys_a.begin(); ait != polys_a.end(); ait++)
		{
			for(std::list<Polygon *>::iterator bit = polys_b.begin(); bit != polys_b.end(); bit++)
			{
				std::list<Polygon *> clipResult;
				Clip_result cr = (*ait)->clip(**bit,Intersect,clipResult);
				// has been clipped. Add the new polygon to the result set
				if(cr == Default)
				{
					total_result = Default;
					result.splice(result.end(),clipResult);
				}
				// an original polygon is the result. Add it to the result set and mark it
				// for non-deletion. For the subject polygon...
				else if(cr == SubjectPolygon)
				{
					if(total_result == Nothing || total_result == SubjectPolygon)
						total_result = SubjectPolygon;
					else
						total_result = Default;

					result.push_back(*ait);
					// perhaps it is already inside
					if(noclear_a.back() != ait)
						noclear_a.push_back(ait);
				}
				// ... and the clip polygon
				else if(cr == ClipPolygon)
				{
					if(total_result == Nothing || total_result == ClipPolygon)
						total_result = ClipPolygon;
					else
						total_result = Default;

					result.push_back(*bit);
					noclear_b.push_back(bit);
				}
			}
		}
		// clear the polygons in polys_a and polys_b (and delete them) except
		// for the polygons we marked
		std::list<Polygon *>::iterator it;
		std::list<std::list<Polygon *>::iterator>::iterator noit;
		for(it = polys_a.begin(), noit = noclear_a.begin(); it != polys_a.end() && noit != noclear_a.end(); ++it)
		{
			if(it == *noit)
				noit++;
			else
				delete *it;
		}
		for(it = polys_b.begin(), noit = noclear_b.begin(); it != polys_b.end() && noit != noclear_b.end(); ++it)
		{
			if(it == *noit)
				noit++;
			else
				delete *it;
		}
		polys_a.clear();
		polys_b.clear();
	}
	// for exclusion
	else if(op == Exclusion)
	{
		std::list<Polygon *> exclusion_result;
		total_result = BothPolygons;

		// put polys_a and polys_b in the exclusion result list
		exclusion_result.splice(exclusion_result.begin(),polys_b);
		std::list<Polygon *>::iterator it_start2 = exclusion_result.begin();
		std::list<Polygon *>::iterator it_end = it_start2;
		exclusion_result.splice(exclusion_result.begin(),polys_a);

		bool clippedSomething = false;
		// on the first run, take advantage of the fact that no polygon of A
		// intersects with another polygon of A. The same with B. So we only need
		// to do the XOR between polygons of A with polygons of B
		bool separate = true;
		do
		{
			clippedSomething = false;
			std::list<Polygon *> interimResult;

			for(std::list<Polygon *>::iterator it = exclusion_result.begin(); it != it_end; )
			{
				bool gotonext = true;

				std::list<Polygon *>::iterator it2(it);
				++it2;
				if(separate)
					it2 = it_start2;

				for(; it2 != exclusion_result.end(); ++it2)
				{
					std::list<Polygon *> clipResult;
					Clip_result cr = (*it)->clip(**it2,Exclusion,clipResult);
					// the polygons don't intersect
					if(cr == BothPolygons)
						continue;
					// the two polygons X'ed each other out or something was clipped
					else if(cr == Nothing || cr == Default)
					{
						total_result = Default;
						clippedSomething = true;
						// remove the original polygons
						delete *it;
						delete *it2;
						// update iterators also
						bool update_start2 = (separate && it2 == it_start2);
						bool update_end = (it2 == it_end);
						std::list<Polygon *>::iterator tmp = exclusion_result.erase(it2); 
						if(update_start2)
							it_start2 = tmp;
						if(update_end)
							it_end = tmp;

						it = exclusion_result.erase(it);
						// something was clipped: put the clip result into the result list
						if(cr == Default)
							interimResult.splice(interimResult.end(),clipResult);
						gotonext = false;
						break;
					}
				}
				if(gotonext) it++;
			}
			it_end = exclusion_result.begin();
			exclusion_result.splice(exclusion_result.begin(),interimResult);
			separate = false;

		} while(clippedSomething);

		if(exclusion_result.empty())
			if(total_result == Default)
				total_result = Nothing;
		result.splice(result.end(),exclusion_result);
	}

	return total_result;
}

Polygon* PolygonClipper::createPolygon(const std::list<PolyVertex>& poly) const
{
	std::vector<Vertex> construct;
	construct.reserve(poly.size());
	for(const_polyvertex_it it = poly.begin(); it != poly.end(); ++it)
	{
		construct.push_back(it->data);
	}
	return new Polygon(construct);
}