/*
 * 2D Polygon Landscape
 *
 * Copyright (c) 2011  Tobias Zwick
 */

#ifndef _POLYGONCLIPPER_H_
#define _POLYGONCLIPPER_H_

#include <list>
#include <map>
#include <set>

#include "Polygon.h"
#include "Line.h"
#include "Vertex.h"

class PolygonClipper
{
public:
	PolygonClipper(const Polygon &subject, const Polygon &clip);
	~PolygonClipper();
	Clip_result clip(Clipping_operation op, std::list<Polygon *>& result);
	// NOTE: these only return something before the clip operation has been called
	std::list<Polygon*> getSubjectPolygons() const;
	std::list<Polygon*> getClipPolygons() const;

private:
	/* +++++++++++++++++++++++++++++ Local Enums +++++++++++++++++++++++++++++ */

	// the different tags for the edges during a clipping operation
	enum Poly_tag { Inside, Outside, ReverseShared, Shared, None };

	/* ++++++++++++++++++++++++++++ Nested Classes +++++++++++++++++++++++++++ */

	// shortcut for iterator to one vertex during clipping process
	class PolyVertex;
	typedef std::list<PolyVertex>::iterator polyvertex_it;
	typedef std::list<PolyVertex>::const_iterator const_polyvertex_it;

	/* PolyVertex

		 This struct is used during the clipping algorithm. Additionally to the
		 vertex data, a link to the PolyVertexLink is stored, the tag for the 
		 successive edge and a (for that polygon) unique ID.
	*/
	class PolyVertexLink;
	class PolyVertex
	{
	public:
		PolyVertex(const Vertex &a, unsigned int i)
			: link(NULL), tag(None), data(a), selflink(NULL), id(i)  {}

		PolyVertexLink* link;
		Poly_tag tag;
		Vertex data;
		PolyVertexLink* selflink;
		unsigned int id;
	};

	struct PolyVertexItComp {
		bool operator() (const polyvertex_it& lhs, const polyvertex_it& rhs) const
		{ return lhs->id < rhs->id; }
	};

	/* PolyVertexLink

		 This helper class is used during the clipping algorithm. It stores
		 each all iterators to vertices that are part of the intersection
		 a polyvertexlink represents. For weakly simple polygons, this can
		 be several iterators in the same polygon.
	*/
	class PolyVertexLink
	{
	public:
		PolyVertexLink():visited(false) {}
		std::set<polyvertex_it,PolyVertexItComp> positionsInA;
		std::set<polyvertex_it,PolyVertexItComp> positionsInB;
		bool visited;
	};

	// shortcut to an iterator for the vertices that are part of an intersection
	typedef std::set<polyvertex_it,PolyVertexItComp>::iterator intersect_pos_it;

	/* PolyPathInfo

		 This helper class is used during the clipping algorithm. It represents
		 a possible path that can be gone during the construction of a polygon.
	*/
	class PolyPathInfo
	{
	public:
		PolyPathInfo(std::list<PolyVertex> &p, polyvertex_it &w, bool r):
				path(p), walker(w), reverse(r) {}
		std::list<PolyVertex> &path;
		polyvertex_it walker;
		bool reverse;
	};

	typedef std::map<Vertex,PolyVertexLink,VertexComp>::iterator polyvertexlink_it;

	/* polygon clipping step 1 */

	bool calculateIntersections();
	bool calculateIntersectionsSimple();

	bool handleIntersection(polyvertex_it &i,	polyvertex_it &j);

	void handleLineIntersection(	const Line& intersectLine, const Line& line,
																polyvertex_it &insert, std::list<PolyVertex> &poly,
																polyvertex_it &pos1, polyvertex_it &pos2);

	void handlePointIntersection(	const Vertex &intersectPoint,	const Line &line,
																polyvertex_it &insert, std::list<PolyVertex> &poly,
																polyvertex_it &pos);

	bool handleSelfIntersection(	polyvertex_it &it, polyvertex_it &jt, bool forB);

	void checkForNewIntersections(	polyvertex_it &posA, polyvertex_it &posB );

	void splitUpAtSelfIntersections(	std::list<PolyVertex> &poly,
																		std::list<Polygon*> &polygons,
																		std::map<Vertex,PolyVertexLink,VertexComp> &self_intersections,
																		bool forB);

	int clearSelfIntersectionLinks(	polyvertex_it &polygon_start,
																	std::map<Vertex,PolyVertexLink,VertexComp> &self_intersections);

	void eliminateHolesSimple(std::list<Polygon *> &polygons) const;

	Clip_result clipMultiple(	Clipping_operation op,
														std::list<Polygon *> &polys_a,
														std::list<Polygon *> &polys_b,
														std::list<Polygon *> &result,
														bool simpleIntersections = false) const;

	/* polygon clipping step 2 */

	Poly_tag tagPolygons();

	Poly_tag tagForIntersection(const polyvertex_it &p,	bool forB);

	double getEnclosingAngle(double prev, double next) const;

	Poly_tag getLineOrientation(const Line &prevLine, const Line &nextLine, const Line &line) const;

	void calculateClosestEdgesToLine(	const Line& line, Line &prevLine, Line &nextLine,
																		std::set<polyvertex_it,PolyVertexItComp> &verticesInOtherPolygon,
																		std::list<PolyVertex> &otherPolygon, bool forB) const;

	/* polygon clipping step 3 */

	void constructPolygons(	Clipping_operation op, std::list<Polygon*> &clipResult);

	std::list<PolyPathInfo> getAllAvailablePaths(	PolyVertexLink* link, Clipping_operation op);

	Clip_result standardClipResult(Clipping_operation op) const;

	Clip_result enclosedClipResult(Clipping_operation op, Poly_tag tag_for_a) const;

	/* eliminating holes */

	void eliminateHoles(std::list<Polygon *> &clipResult) const;

	/* handle multiple intersections */
	Clip_result handleMultipleClippingOperations(Clipping_operation op, std::list<Polygon *> &result);
	Polygon* createPolygon(const std::list<PolyVertex>& poly) const;
	void clearPolygonList(std::list<Polygon *> &polygons) const;

	/* ++++++++++++++++++++++++++ Private variables ++++++++++++++++++++++++++ */

	unsigned int id;
	std::list<PolyVertex> polygon_a;
	std::list<PolyVertex> polygon_b;
	std::list<Polygon*> polygons_of_a;
	std::list<Polygon*> polygons_of_b;
	std::map<Vertex,PolyVertexLink,VertexComp> intersections;
	std::map<Vertex,PolyVertexLink,VertexComp> self_intersections_a;
	std::map<Vertex,PolyVertexLink,VertexComp> self_intersections_b;
	const Polygon &subject_polygon;
	const Polygon &clip_polygon;
	bool has_self_intersections_in_a;
	bool has_self_intersections_in_b;
	bool has_intersections;
	bool outside_bounding_box;
	bool dont_delete_polygonlists;
};

#endif